# coding=utf-8
"""
.. moduleauthor:: Shuhua Gao

This :mod:`copt` module provides advanced routines for numerical constant optimization in gene expression programming.
It is known that both genetic programming (GP) and GEP can usually find a near-optimal solution quite quickly. However,
they both have difficulty in optimizing the numerical constants integrated in the model, especially when the constants
are continuous real numbers. To optimize the constants efficiently, a local optimizer is often incorporated, i.e., GEP
mainly aims to formulate the *true* structure of the model, while the embedded local optimizer is dedicated to improving
the constants existent in the individuals.

To reduce computational cost, such local optimization is applied every *k* generations and often only applied to
selected individuals. Accordingly, two arguments *opt_period* and *opt_selector* are required by the algorithms.

Such kind of algorithms combining evolutionary algorithms and local improvement procedures are often referred to as
`memetic algorithm <https://en.wikipedia.org/wiki/Memetic_algorithm>`_ in literature.

A common interface for a local optimizer is ``optimize(population, selector, generation, **kwargs) -> population``.
"""


def gep_opt(population, toolbox, n_generations=100, n_elites=1,
            stats=None, hall_of_fame=None, verbose=__debug__,
            optimizer=None, opt_period=None, opt_selector=None, **kwargs):
    pass


def _hill_climbing(population, selector, generation, **kwargs):
    pass


def gep_hill_climbing(population, toolbox, n_generations=100, n_elites=1,
            stats=None, hall_of_fame=None, verbose=__debug__,
            opt_period=30, opt_selector='best 10', max_step=1, rnc_gen=None):
    pass


# coding=utf-8
"""
.. moduleauthor:: Shuhua Gao

This :mod:`symbol` module defines the classes encapsulating the symbols in GEP. A gene in GEP is composed of multiple
symbols,  which include *terminals* and *functions*. Together, they are called *primitives*. This module also
provides a :class:`PrimitiveSet` class, which represents a primitive set containing :class:`Terminal`
and :class:`Function` items, for management purpose.

This module implementation refers a lot to the
`deap.gp <http://deap.readthedocs.io/en/master/tutorials/advanced/gp.html>`_ module in DEAP,
especially on primitive management and parsing during evaluation. However, unlike genetic programming (GP) in
:mod:`deap.gp`, in GEP we usually only consider loosely  typed evolution, i.e., the terminals and functions don't
require specific data types. As a result, the codes can be greatly simplified.

.. note::
    In GEP and GP terminology, there are two kinds of primitives: *functions* and *terminals*. In
    `deap.gp <http://deap.readthedocs.io/en/master/tutorials/advanced/gp.html>`_, a *primitive* actually refers to
    a function. To avoid possible ambiguity, in :mod:`geppy` they are explicitly named *function* and *terminal*
    respectively. For instance, the :meth:`PrimitiveSet.add_function` method can be used to add a function primitive.

**Reference:**

.. [FC2006] Ferreira, Cândida. Gene expression programming: mathematical modeling by an artificial
                intelligence. Vol. 21. Springer, 2006.
"""
import keyword
import copy
import numbers


def _is_nonkeyword_identifier(var):
    """
    Check whether the string *var* is a valid non-keyword Python identifier.
    For example, 'ab' and 'x1' satisfy this condition, while 'if' and 'lambda' are keywords.
    :param var: a string
    :return: bool
    """
    if var.isidentifier() and not keyword.iskeyword(var):
        return True
    return False


class Primitive:
    """
    Class that encapsulates a primitive in GEP. A primitive may be a function or a terminal.
    """

    def __init__(self, name, arity):
        """
        Initialize a primitive.

        :param name: str, name of the primitive
        :param arity: int, arity of the primitive
        """
        self._name = name
        self._arity = arity

    @property
    def arity(self):
        """
        Get the arity of this primitive. For a terminal, the arity is 0.
        """
        return self._arity

    @property
    def name(self):
        """
        Get the name of this primitive.
        """
        return self._name

    def format(self, *args):
        """
        Get a string representation of the underlying concept with possible arguments. The derived classes should
        implement this method.

        :param args: variable number of arguments
        :return: str, a string
        """
        raise NotImplementedError()

    def __str__(self):
        return self.name

    def __repr__(self):
        return '{}(name={}, arity={})'.format(self.__class__, self.name, self.arity)

    def __deepcopy__(self, memodict):
        """
        Faster deep copy because generally a primitive is immutable except the ephemeral terminals
        The derived classes should rewrite this method if needed.
        """
        return self


class Function(Primitive):
    """
    Class that encapsulates a function in GEP. Note that this class only stores the function ID, i.e., the *name*
    attribute instead of the callable function itself. On the other hand, the underlying callable is retrieved
    somewhere else when needed, for example, from :class:`PrimitiveSet`. Thus, in the whole GEP program the provided
    name for each function must be unique.
    """

    def __init__(self, name, arity):
        """
        Initialize a function.

        :param name: str, name of the function, must be valid non-keyword Python identifier
        :param arity: int, arity of the function
        """
        assert _is_nonkeyword_identifier(name), \
            'Name of a function must be a valid non-keyword Python identifier'
        super().__init__(name, arity)
        args = ', '.join('{{{}}}'.format(index) for index in range(arity))  # like '{0}, {1}, {2}'
        self._seq = name + '(' + args + ')'  # e.g., add, --> 'add({0}, {1})'

    def format(self, *args):
        """
        Insert the arguments *args* into the function and get a Python statement to call the functions with the
        arguments in a string form. This returned string can afterwards by evaluated using the builtin
        :func:`eval` function.

        >>> f = Function('add', 2)
        >>> f.format(2, 10)
        'add(2, 10)'

        :param args: arguments, whose number should be equal to the arity of this function
        :return: str, a string form of function calling with arguments
        """
        assert len(args) == self.arity, "Function {} requires {} arguments while {} are provided.".format(
            self.name, self.arity, len(*args))
        return self._seq.format(*args)


class Terminal(Primitive):
    """
    Class that encapsulates a terminal in GEP.
    """

    def __init__(self, name, value):
        """
        Initialize a terminal.

        :param name: str, name of the terminal
        :param value: value of the terminal
        """
        super().__init__(name, arity=0)
        self._value = value

    def format(self):
        """
        Get a string representation of this terminal for subsequent evaluation purpose, which may be a symbol or a
        numeric string.

        :return: str, a string representation
        """
        return self.name

    @property
    def value(self):
        """
        Get the value of this terminal.
        """
        return self._value

    def __repr__(self):
        return '{}(name={}, value={})'.format(self.__class__, self.name, self.value)

    def __str__(self):
        return self.format()


class ConstantTerminal(Terminal):
    """
    Class that represents a constant terminal, whose value will never change during the whole evolution.
    The value of a constant terminal can only be a literal variable , such as a numeric number or a Boolean variable.
    """
    def __init__(self, value):
        """
        Initialize a constant terminal with the given *value*.

        :param value: int, float, bool, value of the constant, can only be a number or a Boolean variable
        """
        assert isinstance(value, numbers.Number) or isinstance(value, bool), \
            'The value of a constant terminal can only be a number or a bool variable.'
        super().__init__(repr(value), value)


class SymbolTerminal(Terminal):
    """
    Class that represents a symbolic terminal. Only the :attr:`name` of a symbol terminal participates in the genome
    construction, while its value is retrieved from the :class:`PrimitiveSet` only during evaluation. Therefore, the
    :attr:`name` of a symbol terminal must be a valid and unique Python identifier.
    The value of a symbol terminal can be any object that leads to a legal expression when evaluated.
    """
    def __init__(self, name):
        """
        Initialize a symbol terminal with the symbol given by *name*.

        :param name: str, must be a valid non-keyword Python identifier
        """
        assert _is_nonkeyword_identifier(name), \
            'Name of a symbol terminal must be a valid non-keyword Python identifier'
        super().__init__(name, value=None)


class EphemeralTerminal(Terminal):
    """
    Class that encapsulates an ephemeral numeric constant terminal in GEP, which can be used to build a normal constant
    terminal with a randomly produced value when needed using the :meth:`Ephemeral.generate` method.

    Just as the name implies, the value of an ephemeral terminal may be changed during evolution, either by mutation or
    by some local search heuristics, to optimize the numerical constants in a mathematical model.

    .. note::
            This special terminal named *ephemeral random constant* was originally introduced in genetic programming
            by Koza to handle numerical constants in evolutionary programming. Another way to evolve simple constants in
            GEP is to add a RNC domain in genes and use the GEP-RNC algorithm. See Chapter 5 of [FC2006]_.

    .. seealso::
        :func:`~geppy.tools.mutation.mutate_uniform_ephemeral` for ephemeral mutation, and
        :class:`~geppy.core.entity.GeneDc` for the GEP-RNC algorithm.
    """

    def __init__(self, name, gen):
        """
        Initialize an ephemeral terminal with the given *name* and the random number generator *gen*. Its initial value
        is randomly generated by *gen* and can later be set explicitly with the :attr:`value` property. Alternatively,
        its value can be updated to a new value generated by *gen* with the :meth:`update_value` method. Note that just
        like a constant terminal, the value of an ephemeral terminal can only be a number or a bool.

        :param name: str, name of the ephemeral constant
        :param gen: callable, an ephemeral number generator, which should return a random value when called with
            no arguments ``gen()``.
        """
        super().__init__(name, value=None)
        self._gen = gen
        self.value = self._gen()

    @property
    def generator(self):
        """
        Get the ephemeral random number generator.
        """
        return self._gen

    @property
    def value(self):
        """
        Get the value.
        """
        return self._value

    @value.setter
    def value(self, val):
        """
        Set the value to be *val*.
        """
        assert isinstance(val, numbers.Number) or isinstance(val, bool), \
            'The value of a constant terminal can only be a number or a bool variable.'
        self._value = val

    def format(self):
        """
        Get a string representation of this terminal for subsequent evaluation purpose, which is equivalently
        ``repr(self.value)``.
        """
        return repr(self.value)

    def update_value(self):
        """
        Update the value of this ephemeral constant in place, whose new value is randomly produced by the random number
        generator :attr:`generator`.
        """
        self._value = self._gen()

    def __repr__(self):
        return '{}(name={}, gen={})'.format(self.__class__, self.name, self.generator)

    def __deepcopy__(self, memodict={}):
        """
        Deep copy the ephemeral and returns a new ephemeral terminal with the same value.
        """
        return copy.copy(self)


class RNCTerminal(Terminal):
    """
    A special terminal, which is just a placeholder representing a random numerical constant (RNC) in the GEP-RNC
    algorithm. This class is mainly used internally by the :class:`~geppy.core.entity.GeneDc` class. The name of a
    :class:`RNCTerminal` object is '?' by default, and its value is retrieved dynamically according to the GEP-RNC
    algorithm. Refer to Chapter 5 of [FC2006]_ for more details.

    In *geppy* and the GEP-RNC algorithm, the RNC terminal is just a placeholder and the default name '?' is
    recommended.
    """

    def __init__(self, name='?'):
        """
        Initialize a RNC terminal.

        :param name: str, default '?', name of the terminal
        """
        super().__init__(name, value=None)


class PrimitiveSet:
    """
    A class representing a primitive set, which contains the primitives (terminals and functions) that are used in GEP.

    .. note::
        Each function :class:`Function` and symbol terminal :class:`SymbolTerminal`
        must have their unique names. This is because internally their true value (or
        the callable for a function) will be be stored in a dictionary :attr:`globals` inside :class:`PrimitiveSet` and
        later be retrieved when compiling a genome. Consequently, the name must be a valid non-keyword Python identifier.
        To learn more about the underlying mechanism, refer to the Python builtin function :func:`compile`.

    .. note::
        To use the GEP-RNC algorithm, i.e., to use :class:`~geppy.core.entity.GeneDc` for numerical constant handling,
        please use the :meth:`PrimitiveSet.add_rnc_terminal` method, which will add a special terminal
        of type :class:`~geppy.core.symbol.RNCTerminal` internally. Then, use a chromosome composed of
        :class:`~geppy.core.entity.GeneDc` genes as the individual.
        Refer to Chapter 5 of [FC2006]_ for more details.
    """

    def __init__(self, name, input_names):
        """
        Initiate a primitive set with the given *name* and the list of input names.

        :param name: name of the primitive set
        :param input_names: iterable, a list of names for the inputs in the GEP problem, for instance,
            ``['x', 'y']`` for two inputs. Internally, a symbol terminal is built for each input.
        """
        self._name = name
        self._functions = []
        self._terminals = []
        self._globals = {'__builtins__': None}
        self._inputs = input_names
        self._unique_names = set()
        # add input terminals
        for name in input_names:
            self._assert_name_unique(name)
            self.add_symbol_terminal(name, value='input')

    def add_function(self, func, arity, name=None):
        """
        Add a function, which is internally encapsulated as a :class:`Function` object.

        :param func: callable
        :param arity: number of arguments accepted by *func*
        :param name: name of *func*, default ``None``.
            If remaining ``None``, then the ``func.__name__`` attribute is used instead.
        """
        if name is None:
            name = func.__name__
        self._assert_name_unique(name)
        function_ = Function(name, arity)
        self._functions.append(function_)
        self._globals[name] = func

    def add_constant_terminal(self, value):
        """
        Add a terminal which is a constant.

        There is no need to store such kind of terminals into *globals* for later
        evaluation, because their string representation acquired by :meth:`~Terminal.format` can be used directly.

        :param value: value of the terminal. Only numeric and Boolean types can be accepted.
        """
        self._terminals.append(ConstantTerminal(value))

    def add_symbol_terminal(self, name, value):
        """
        Add a symbolic terminal, whose name points to the true value stored in :attr:`globals`.

        For example, we may add a terminal for the constant *pi*, with ``add_symbol_terminal(3.1415, 'pi')``.

        :param name: name of the terminal
        :param value: value of the terminal
        """
        self._assert_name_unique(name)
        t = SymbolTerminal(name)
        self._globals[name] = value
        self._terminals.append(t)

    def _assert_name_unique(self, name):
        """
        Assert whether a *name* is a valid symbol in this set. It should be (1) unique in *globals* and (2) a valid
        Python identifier. If the supplied *name* is acceptable, it is added to the :attr:`~PrimitiveSet._unique_names`
        attribute.
        """
        assert name not in self._globals, "A primitive with the name '{}' already exists.".format(name)

    def add_ephemeral_terminal(self, name, gen):
        """
        Add an ephemeral constant to the set. An ephemeral's true value is generated by *gen*, which should be a no
        argument function that returns a random value, when the ephemeral is picked to build a gene. Once an ephemeral
        is picked, its :meth:`Ephemeral.generate` method is called to produce a random value.

        In GEP, the ephemeral terminals (i.e., their values) can be evolved by simple mutation or advanced
        local search optimization.

        :param name: str, name of the ephemeral
        :param gen: callable, a random number generator which returns a random constant when called by ``gen()``

        Usually calling :meth:`add_ephemeral_terminal` once is sufficient to generate the numerical coefficients in the
        model, because during evolution multiple independent copies of the ephemeral terminal will be built with
        various values.
        Nevertheless, if the numerical coefficients have various ranges, it may be more efficient to add multiple
        ephemeral terminals.

        .. note::
            The special terminal named *ephemeral random constant* was originally introduced in genetic programming
            by Koza to handle numerical constants in evolutionary programming. Alternatively, to evolve the numerical
            constants in GEP, we can also use the GEP-RNC algorithm.
            See Chapter 5 of [FC2006]_.

        .. seealso::
            :func:`~geppy.tools.mutation.mutate_uniform_ephemeral` for ephemeral mutation, and
            :class:`~geppy.core.entity.GeneDc` for the GEP-RNC algorithm.
        """
        self._assert_name_unique(name)
        self._terminals.append(EphemeralTerminal(name, gen))

    def add_rnc_terminal(self, name='?'):
        """
        Add a special terminal representing a random numerical constant (RNC), as defined in the GEP-RNC algorithm.
        This terminal's value is retrieved dynamically from an RNC array attached to a gene of type
        :class:`~geppy.core.entity.GeneDc` according to the GEP-RNC algorithm.
        See also :class:`RNCTerminal` and refer to Chapter 5 of [FC2006]_ about GEP-RNC.

        :param name: str, name of the terminal. For a RNC terminal, generally there is no need to specify a name and
            the default value '?' is recommended.

        Usually it is sufficient to call this method once to add only one RNC terminal, since the values of the
        RNC terminals at different positions are different random numerical constants..
        """
        self._terminals.append(RNCTerminal(name))

    @property
    def functions(self):
        """
        Get all the functions.
        """
        return self._functions

    @property
    def terminals(self):
        """
        Get all the terminals.
        """
        return self._terminals

    @property
    def ephemerals(self):
        """
        Get all the ephemeral terminals in this set.
        """
        return [t for t in self.terminals if isinstance(t, EphemeralTerminal)]

    @property
    def name(self):
        """
        Get the name of this primitive set.
        """
        return self._name

    @property
    def input_names(self):
        """
        Get a list of names for the input arguments.
        """
        return self._inputs

    def __str__(self):
        """
        Gets an overview of the functions and terminals in this primitive set.
        """
        terminals = '[{}]'.format(', '.join(str(t) for t in self.terminals))
        functions = '[{}]'.format(', '.join(str(f) for f in self.functions))
        return 'PrimitiveSet {}\n\tFunctions: {}\n\tTerminals: {}'.format(self.name, functions, terminals)

    @property
    def globals(self):
        """
        Get a dictionary which can be used to set up the evaluation/execution environment. This dictionary can be
        fed into the builtin :func:`eval` and :func:`exec` functions for expression evaluation.

        For example, if we call ``add_function(max, 2, 'max2')``, then ``globals['max2']`` corresponds to a
        :class:`Function` object encapsulating the :func:`max` function.
        """
        return self._globals

    @property
    def max_arity(self):
        """
        Get the max arity of functions in this primitive set.
        """
        return max(f.arity for f in self.functions)



# coding=utf-8
"""
.. moduleauthor:: Shuhua Gao

The module :mod:`entity` defines the core data structures used in GEP, including the gene, the chromosome, the
K-expression and the expression tree. We use primitives (functions and terminals) to build genes and then compose
a chromosome with one or multiple genes. Refer to the :mod:`geppy.core.symbol` module for information on primitives.

A chromosome composed of only one gene is called a *monogenic* chromosome, while one composed of more than one genes
is named a *multigenic* chromosome. A multigenic chromosome can be assigned a linking function to combine the results
from multiple genes into a single result.
"""
import copy
from ..tools.generator import *
from ..core.symbol import Function, RNCTerminal, Terminal


_DEBUG = False


class KExpression(list):
    """
    Class representing the K-expression, or the open reading frame (ORF), of a gene in GEP. A K-expression is actually
    a linear form of an expression tree obtained by level-order traversal.
    """
    def __init__(self, content):
        """
        Initialize a K-expression.

        :param content: iterable, each element is a primitive (function or terminal)
        """
        list.__init__(self, content)

    def __str__(self):
        """
        Get a string representation of this expression by joining the name of each primitive.
        :return:
        """
        return '[' + ', '.join(ele.name for ele in self) + ']'

    def __repr__(self):
        return str(self.__class__) + '[{}]'.format(', '.join(repr(p) for p in self))


class ExpressionTree:
    """
    Class representing an expression tree (ET) in GEP, which may be obtained by translating a K-expression, a
    gene, or a chromosome, i.e., genotype-phenotype mapping.
    """
    def __init__(self, root):
        """
        Initialize a tree with the given *root* node.

        :param root: :class:`ExpressionTree.Node`, the root node
        """
        self._root = root

    @property
    def root(self):
        """
        Get the root node of this expression tree.
        """
        return self._root

    class Node:
        """
        Class representing a node in the expression tree. Each node has a variable number of children, depending on
        the arity of the primitive at this node.
        """
        def __init__(self, name):
            self._children = []
            self._name = name

        @property
        def children(self):
            """
            Get the children of this node.
            """
            return self._children

        @property
        def name(self):
            """
            Get the name (label) of this node.
            """
            return self._name

    @classmethod
    def from_genotype(cls, genome):
        """
        Create an expression tree by translating *genome*, which may be a K-expression, a gene, or a chromosome.

        :param genome: :class:`KExpression`, :class:`Gene`, or :class:`Chromosome`, the genotype of an individual
        :return: :class:`ExpressionTree`, an expression tree
        """
        if isinstance(genome, Gene):
            return cls._from_kexpression(genome.kexpression)
        elif isinstance(genome, KExpression):
            return cls._from_kexpression(genome)
        elif isinstance(genome, Chromosome):
            if len(genome) == 1:
                return cls._from_kexpression(genome[0].kexpression)
            sub_trees = [cls._from_kexpression(gene.kexpression) for gene in genome]
            # combine the sub_trees with the linking function
            root = cls.Node(genome.linker.__name__)
            root.children[:] = sub_trees
            return cls(root)
        raise TypeError('Only an argument of type KExpression, Gene, and Chromosome is acceptable. The provided '
                        'genome type is {}.'.format(type(genome)))

    @classmethod
    def _from_kexpression(cls, expr):
        """
        Create an expression tree from a K-expression.

        :param expr: a K-expression
        :return: :class:`ExpressionTree`, an expression tree
        """
        if len(expr) == 0:
            return None
        # first build a node for each primitive
        nodes = [cls.Node(p.name) for p in expr]
        # connect each node to its children if any
        i = 0
        j = 0
        while i < len(nodes):
            for _ in range(expr[i].arity):
                j += 1
                nodes[i].children.append(nodes[j])
            i += 1
        return cls(nodes[0])


class Gene(list):
    """
    A single gene in GEP, which is a fixed-length linear sequence composed of functions and terminals.
    """
    def __init__(self, pset, head_length):
        """
        Instantiate a gene.
        :param head_length: length of the head domain
        :param pset: a primitive set including functions and terminals for genome construction.

        Supposing the maximum arity of functions in *pset* is *max_arity*, then the tail length is automatically
        determined to be ``tail_length = head_length * (max_arity - 1) + 1``. The genome, i.e., list of symbols in
        the instantiated gene is formed randomly from *pset*.
        """
        self._head_length = head_length
        genome = generate_genome(pset, head_length)
        list.__init__(self, genome)

    @classmethod
    def from_genome(cls, genome, head_length):
        """
        Build a gene directly from the given *genome*.

        :param genome: iterable, a list of symbols representing functions and terminals
        :param head_length: length of the head domain
        :return: :class:`Gene`, a gene
        """
        g = super().__new__(cls)
        super().__init__(g, genome)
        g._head_length = head_length
        return g

    def __str__(self):
        """
        Return the expression in a human readable string, which is also a legal python code that can be evaluated.

        :return: string form of the expression
        """
        expr = self.kexpression
        i = len(expr) - 1
        while i >= 0:
            if expr[i].arity > 0:  # a function
                f = expr[i]
                args = []
                for _ in range(f.arity):
                    ele = expr.pop()
                    if isinstance(ele, str):
                        args.append(ele)
                    else:  # a terminal or an ephemeral class
                        args.append(ele.format())
                expr[i] = f.format(*reversed(args))  # replace the operator with its result (str)
            i -= 1

        # the final result is at the root. It may happen that the K-expression contains only one terminal at its root.
        return expr[0] if isinstance(expr[0], str) else expr[0].format()

    @property
    def kexpression(self):
        """Get the K-expression of type :class:`KExpression` represented by this gene.
        """
        # get the level-order K expression
        expr = KExpression([self[0]])
        i = 0
        j = 1
        while i < len(expr):
            for _ in range(expr[i].arity):
                expr.append(self[j])
                j += 1
            i += 1
        return expr

    @property
    def head_length(self):
        """
        Get the length of the head domain.
        """
        return self._head_length

    @property
    def tail_length(self):
        """
        Get the length of the tail domain.
        """
        return len(self) - self.head_length

    @property
    def max_arity(self):
        """
        Get the max arity of the functions in the primitive set with which the gene is built.
        """
        return (len(self) - 1) // self.head_length

    @property
    def head(self):
        """
        Get the head domain of the gene.
        """
        return self[0: self.head_length]

    @property
    def tail(self):
        """
        Get the tail domain of the gene.
        """
        return self[self.head_length: self.head_length + self.tail_length]

    def __repr__(self):
        """
        Return a string representation of a gene including all the symbols. Mainly for debugging purpose.

        :return: a string
        """
        info = [str(p) for p in self]
        return '{} ['.format(self.__class__) + ', '.join(info) + ']'


class GeneDc(Gene):
    """
    Class represents a gene with an additional Dc domain to handle numerical constants in GEP.

    The basic :class:`Gene`
    has two domains, a head and a tail, while this :class:`GeneDc` class introduces another domain called *Dc*
    after the tail. The length of the Dc domain is equal to the length of the tail domain. The Dc domain only stores
    the indices of numbers present in a separate array :meth:`rnc_array`, which collects a group of candidate
    random numerical constants (RNCs). Thus, each :class:`GeneDc` instance comes with a *rnc_array*, which are
    generally different among different instances.

    In addition to the operators of the basic gene expression algorithm (mutation, inversion, transposition, and
    recombination), Dc-specific operators in GEP-RNC are also created to better evolve the Dc domain and to manipulate
    the attached set of constants *rnc_array*. Such operators are all suffixed with '_dc' in
    :mod:`~geppy.tools.crossover` and :mod:`~geppy.tools.mutation` modules, for example, the method
    :func:`~geppy.tools.mutation.invert_dc`.

    The *rnc_array* associated with each :class:`GeneDc` instance can be provided explicitly when creating
    a :class:`GeneDc` instance. See :meth:`from_genome`. Or more generally, a
    random number generator *rnc_gen* can be provided, with which a specified number of RNCs are generated during the
    creation of gene instances.

    A special terminal of type :class:`~geppy.core.symbol.TerminalRNC` is used internally to represent the RNCs.

    .. note::
        To create an effective :class:`GeneDc` gene, the primitive set should contain at least one
        :class:`~geppy.core.symbol.TerminalRNC` terminal. See :meth:`~geppy.core.symbol.PrimitiveSet.add_rnc`
        for details.

    Refer to Chapter 5 of [FC2006]_ for more knowledge about GEP-RNC.
    """
    def __init__(self, pset, head_length, rnc_gen, rnc_array_length):
        """
        Initialize a gene with a Dc domain.

        :param head_length: length of the head domain
        :param pset: a primitive set including functions and terminals for genome construction.
        :param rnc_gen: callable, which should generate a random number when called by ``rnc_gen()``.
        :param rnc_array_length: int, number of random numerical constant candidates associated with this gene,
            usually 10 is enough

        Supposing the maximum arity of functions in *pset* is *max_arity*, then the tail length is automatically
        determined to be ``tail_length = head_length * (max_arity - 1) + 1``. The genome, i.e., list of symbols in
        the instantiated gene is formed randomly from *pset*. The length of Dc domain is equal to *tail_length*, i.e.,
        the tail domain and the Dc domain share the same length.
        """
        # first generate the gene without Dc
        super().__init__(pset, head_length)
        t = head_length * (pset.max_arity - 1) + 1
        d = t
        # complement it with a Dc domain
        self._rnc_gen = rnc_gen
        dc = generate_dc(rnc_array_length, dc_length=d)
        self.extend(dc)
        # generate the rnc array
        self._rnc_array = [self._rnc_gen() for _ in range(rnc_array_length)]

    @classmethod
    def from_genome(cls, genome, head_length, rnc_array):
        """
        Build a gene directly from the given *genome* and the random numerical constant (RNC) array *rnc_array*.

        :param genome: iterable, a list of symbols representing functions and terminals (especially the special RNC
            terminal of type :class:`~geppy.core.symbol.TerminalRNC`). This genome should have three
            domains: head, tail and Dc. The Dc domain should be composed only of integers representing indices into the
            *rnc_array*.
        :param head_length: length of the head domain. The length of the tail and Dc domain should follow the rule and
            can be determined from the head_length: ``dc_length = tail_length = (len(genome) - head_length) / 2``.
        :param rnc_array: the RNC array associated with the gene, which contains random constant candidates
        :return: :class:`GeneDc`, a gene
        """
        g = super().from_genome(genome, head_length)  # the genome is copied and the head-length is fixed
        g._rnc_array = copy.deepcopy(rnc_array)
        return g

    @property
    def dc_length(self):
        """
        Get the length of the Dc domain, which is equal to :meth:`GeneDc.tail_length`
        """
        return self.tail_length

    @property
    def rnc_array(self):
        """
        Get the random numerical array (RNC) associated with this gene.
        """
        return self._rnc_array

    @property
    def tail_length(self):
        """
        Get the tail domain length.
        """
        return (len(self) - self.head_length) // 2

    @property
    def max_arity(self):
        """
        Get the max arity of functions in the primitive set used to build this gene.
        """
        # determine from the head-tail-Dc length relations
        return (len(self) - 2 + self.head_length) // (2 * self.head_length)

    @property
    def dc(self):
        """
        Get the Dc domain of this gene.
        :return:
        """
        return self[self.head_length + self.tail_length: self.head_length + self.tail_length + self.dc_length]

    # def __deepcopy__(self, memodict):
    #     return self.__class__.from_genome(self, self.head_length, self.rnc_array)

    def __str__(self):
        """
        Return the expression in a human readable string, which is also a legal python code that can be evaluated.
        The special terminal representing a RNC will be replaced by their true values retrieved from the array
        :meth:`rnc_array`.

        :return: string form of the expression
        """
        expr = self.kexpression
        n_total_rncs = sum(1 if isinstance(p, RNCTerminal) else 0 for p in expr)  # how many RNCs in total?
        n_encountered_rncs = 0  # how many RNCs we have encountered in reverse order?
        i = len(expr) - 1
        while i >= 0:
            if expr[i].arity > 0:  # a function
                f = expr[i]
                args = []
                for _ in range(f.arity):
                    ele = expr.pop()
                    if isinstance(ele, str):
                        args.append(ele)
                    else:  # a terminal or a RNC terminal
                        if isinstance(ele, RNCTerminal):
                            # retrieve its true value
                            which = n_total_rncs - n_encountered_rncs - 1
                            index = self.dc[which]
                            value = self.rnc_array[index]
                            args.append(str(value))
                            n_encountered_rncs += 1
                        else:   # a normal terminal
                            args.append(ele.format())
                expr[i] = f.format(*reversed(args))  # replace the operator with its result (str)
            i -= 1

        # the final result is at the root
        if isinstance(expr[0], str):
            return expr[0]
        if isinstance(expr[0], RNCTerminal):  # only contains a single RNC, let's retrieve its value
            return str(self.rnc_array[self.dc[0]])
        return expr[0].format()    # only contains a normal terminal

    @property
    def kexpression(self):
        """
        Get the K-expression of type :class:`KExpression` represented by this gene. The involved RNC terminal will be
        replaced by a constant terminal with its value retrived from the :meth:`rnc_array` according to the GEP-RNC
        algorithm.
        """
        n_rnc = 0

        def convert_RNC(p):
            nonlocal n_rnc
            if isinstance(p, RNCTerminal):
                index = self.dc[n_rnc]
                value = self.rnc_array[index]
                n_rnc += 1
                t = Terminal(str(value), value)
                return t
            return p

        # level-order
        expr = KExpression([convert_RNC(self[0])])
        i = 0
        j = 1
        while i < len(expr):
            for _ in range(expr[i].arity):
                expr.append(convert_RNC(self[j]))
                j += 1
            i += 1
        return expr

    def __repr__(self):
        return super().__repr__() + ', rnc_array=[' + ', '.join(str(num) for num in self.rnc_array) + ']'


class Chromosome(list):
    """
    Individual representation in gene expression programming (GEP), which may contain one or more genes. Note that in a
    multigenic chromosome, all genes must share the same head length and the same primitive set. Each element of
    :class:`Chromosome` is of type :class:`Gene`.
    """

    def __init__(self, gene_gen, n_genes, linker=None):
        """
        Initialize an individual in GEP.

        :param gene_gen: callable, a gene generator, i.e., ``gene_gen()`` should return a gene
        :param n_genes: number of genes
        :param linker: callable, a linking function

        .. note:
            If a linker is specified, then it must accept *n_genes* arguments, each produced by one gene. If the
            *linker* parameter is the default value ``None``, then for a monogenic chromosome, no linking is applied,
            while for a multigenic chromosome, the ``tuple`` function is used.
        """
        list.__init__(self, (gene_gen() for _ in range(n_genes)))
        self._linker = linker

    @classmethod
    def from_genes(cls, genes, linker=None):
        """
        Create a chromosome instance by providing a list of genes directly. The number of genes is implicitly specified
        by ``len(genes)``.

        :param genes: iterable, a list of genes
        :param linker: callable, a linking function. Refer to :meth:`Chromosome.__init__` for more details.
        :return: :class:`Chromosome`, a chromosome

        .. caution:
            Every gene of type :class:`Gene` in *genes* should be an independent object. That is, for any two genes
            *g1* and *g2*, the assertion ``g1 is g1`` should return ``False``. Otherwise, unexpected results may happen
            in following evolution.

        """
        c = super().__new__(cls)
        super().__init__(c, genes)
        c._linker = linker
        return c

    @property
    def head_length(self):
        """
        Get the length of the head domain. All genes in a chromosome should have the same head length.
        """
        return self[0].head_length

    @property
    def tail_length(self):
        """
        Get the length of the tail domain. All genes in a chromosome should have the same tail length.
        """
        return self[0].tail_length

    @property
    def max_arity(self):
        """
        Get the max arity of the functions in the primitive set with which all genes are built. Note that all genes
        in a chromosome should share the same primitive set.
        """
        return self[0].max_arity

    def __str__(self):
        """
        Return the expressions in a human readable string.
        """
        if self.linker is not None:
            return '{0}(\n\t{1}\n)'.format(self.linker.__name__, ',\n\t'.join(str(g) for g in self))
        else:
            assert len(self) == 1
            return str(self[0])

    @property
    def kexpressions(self):
        """
        Get a list of K-expressions for all the genes in this chromosome.
        """
        return [g.kexpression for g in self]

    @property
    def linker(self):
        """
        Get the linking function.
        """
        if self._linker is None and len(self) > 1:
            return tuple
        return self._linker

    def __repr__(self):
        """
        Return a string representation of the chromosome including all the symbols in all the genes.
        Mainly for debugging purpose.

        :return: a string
        """
        return '{}[\n\t'.format(self.__class__) + ',\n\t'.join(repr(g) for g in self) + \
               '\n], linker={}'.format(self.linker)





# coding=utf-8
"""
.. moduleauthor:: Shuhua Gao

This module :mod:`visualization` provides utility functions to visualization the expression tree from a given
K-expression, a gene or a chromosome in GEP.
"""

from ..core.entity import KExpression, Chromosome, Gene
from ..core.symbol import Function, Terminal


def _graph_kexpression(expr, starting_index):
    """
    Create a graph for a K-expression *expr* with the node's number starting from *starting_index*.

    :param expr: k-expression
    :param starting_index: the first number of nodes in the expression tree
    :return: A node list, an edge list, and a dictionary of labels.
    """
    assert len(expr) > 0
    nodes = [starting_index + i for i in range(len(expr))]
    edges = []
    labels = {}
    for i, p in enumerate(expr):
        if isinstance(p, Function):
            labels[starting_index + i] = p.name
        elif isinstance(p, Terminal):
            labels[starting_index + i] = p.format()
        else:
            raise RuntimeError('Unrecognized symbol. Normally, a symbol in the K-expression is either a function '
                               'or a terminal')
    i = 0
    j = 0
    while i < len(expr):
        for _ in range(expr[i].arity):
            j += 1
            edges.append((i + starting_index, j + starting_index))
        i += 1
    return nodes, edges, labels


def graph(genome, label_renaming_map=None):
    """
    Construct the graph of a genome. It returns in order a node list, an edge list, and a dictionary of the per node
    labels. The node are represented by numbers, the edges are tuples connecting two nodes (number), and the labels are
    values of a dictionary for which keys are the node numbers.

    :param genome: :class:`~geppy.core.entity.KExpression`, :class:`~geppy.core.entity.Gene`, or
        :class:`~geppy.core.entity.Chromosome`, the genotype of an individual
    :param label_renaming_map: dict, which maps the old name of a primitive (or a linking function)
        to a new one for better visualization. The default label for each node is just the name of the primitive
        placed on this node. For example, you may provide ``renamed_labels={'and_': 'and'}``.
    :return: A node list, an edge list, and a dictionary of labels.

    You can visualize a genome and export the tree visualization to an image file directly using the
    :func:`export_expression_tree` function.
    """
    nodes = []
    edges = []
    labels = {}
    if isinstance(genome, KExpression):
        nodes, edges, labels = _graph_kexpression(genome, 0)
    elif isinstance(genome, Gene):
        nodes, edges, labels = _graph_kexpression(genome.kexpression, 0)
    elif isinstance(genome, Chromosome):
        if len(genome) == 1:
            nodes, edges, labels = _graph_kexpression(genome[0].kexpression, 0)
        else:   # multigenic chromosome, we need to concatenate multiple trees
            starting_index = 1
            sub_roots = []
            for gene in genome:
                expr = gene.kexpression
                sub_roots.append(starting_index)
                sub_nodes, sub_edges, sub_labels = _graph_kexpression(
                    expr, starting_index)
                nodes.extend(sub_nodes)
                edges.extend(sub_edges)
                labels.update(sub_labels)
                starting_index += len(expr)
            # connect subtrees by inserting the linker node as 0
            nodes.append(0)
            for root in sub_roots:
                edges.append((0, root))
            labels[0] = genome.linker.__name__
    else:
        raise TypeError('Only an argument of type KExpression, Gene, and Chromosome is acceptable. The provided '
                        'genome type is {}.'.format(type(genome)))
    # rename_labels labels
    if label_renaming_map is not None:
        for k, v in labels.items():
            if v in label_renaming_map:
                labels[k] = label_renaming_map[v]
    return nodes, edges, labels


def export_expression_tree(genome, label_renaming_map=None, file='tree.png'):
    """
    Construct the graph of a *genome* and then export it to a *file*.

    :param genome: :class:`~geppy.core.entity.KExpression`, :class:`~geppy.core.entity.Gene`, or
        :class:`~geppy.core.entity.Chromosome`, the genotype of an individual
    :param label_renaming_map: dict, which maps the old name of a primitive (or a linking function)
        to a new one for better visualization. The default label for each node is just the name of the primitive
        placed on this node. For example, you may provide ``renamed_labels={'and_': 'and'}``.
    :param file: str, the file path to draw the expression tree, which may be a relative or absolute one.
        If no extension is included in *file*, then the default extension 'png' is used.

    .. note::
        This function currently depends on the :mod:`graphviz` module to render the tree. Please first install the
        `graphviz <https://pypi.org/project/graphviz/>`_ module before using this function.
        Alternatively, you can always obtain the raw graph data with the :func:`graph` function, then postprocess the
        data and render them with other tools as you want.
    """
    import graphviz as gv
    import os.path

    _, edges, labels = graph(genome, label_renaming_map)
    file_name, ext = os.path.splitext(file)
    ext = ext.lstrip('.')
    g = gv.Graph(format=ext)
    for name, label in labels.items():
        g.node(str(name), str(label))  # add node
    for name1, name2 in edges:
        g.edge(str(name1), str(name2))  # add edge
    g.render(file_name)


__all__ = ['graph', 'export_expression_tree']



# coding=utf-8
"""
.. moduleauthor:: Shuhua Gao

The module :mod:`toolbox` mainly provides a class :class:`Toolbox` on the basis of :class:`deap.base.Toolbox` for GEP
operator registration, which is then extensively used in the builtin GEP algorithms. For user-defined custom algorithms,
they should also accept a toolbox instance and retrieve the operators and their associated probability from the toolbox
instance.
"""

from deap import base
import collections


class Toolbox(base.Toolbox):
    """
    A toolbox for evolution that contains the evolutionary operators.
    Initially, the toolbox contains a :meth:`~Toolbox.clone` method that
    duplicates any element it is passed as argument, this method defaults to
    the :func:`copy.deepcopy` function. and a :meth:`map`
    method that applies the function given as first argument to every items
    of the iterables given as next arguments, this method defaults to the
    :func:`map` function. You may populate the toolbox with any other
    function by using the :meth:`~Toolbox.register` method.

    As an extension of :class:`deap.base.Toolbox`, this class adds a :attr:`~Toolbox.pbs` property to specify the
    probabilities for the registered operators. For example, `pbs['mut_uniform']` gives the probability for an
    operator with the alias `'mut_uniform'`. Besides, :attr:`Toolbox.pbs` is an :class:`~collections.OrderedDict` instance
    which can remember the order of entry insertion. Thus, if `'mut_A'` is inserted into :attr:`~Toolbox.pbs` before
    `'mut_B'`, then the operator with the alias `'mut_A'` is applied earlier than the one corresponding to `'mut_B'`.

    A short way to combine operator registration and probability specification together is to pass a keyword-only argument
    `pb` into the :meth:`register` method, which will insert the probability automatically into :attr:`pbs`.
    """

    def __init__(self):
        """
        Initialize an *empty* toolbox with only the default :meth:`clone` and :meth:`map` methods.
        """
        super().__init__()
        self._pbs = collections.OrderedDict()

    def register(self, alias, function, *args, **kargs):
        """
        Register a *function* in the toolbox under the name *alias*. You
        may provide default arguments that will be passed automatically when
        calling the registered function. Fixed arguments can then be overriden
        at function call time.

        :param alias: The name the operator will take in the toolbox. If the
                              alias already exists it will overwrite the the operator
                              already present.
        :param function: The function to which the alias refers.
        :param args: one or more positional arguments to pass to the registered function, optional
        :param kargs: one or more keyword arguments to pass to the registered function, optional

        .. hint::
            Under the hood lies the partial function binding. Check :func:`functools.partial` for details.

        .. note::
            If an operator needs its probability specified, like mutation and crossover operators, it can be done by
            inserting the probability into the :attr:`pbs` dictionary with the same alias. Alternatively, it can be
            given with a the special keyword argument `pb` in this method ::

                tb = Toolbox()
                tb.register('mut_uniform', mutate_uniform, ind_pb=0.02)
                tb.pbs['mut_uniform'] = 0.1

            or equivalently ::

                tb = Toolbox()
                tb.register('mut_uniform', mutate_uniform, ind_pb=0.02, pb=0.1)

            As a result, the special keyword argument `pb` is always excluded from binding into *function*.
        """
        pb = None
        if 'pb' in kargs:
            pb = kargs['pb']
            del kargs['pb']
        super().register(alias, function, *args, **kargs)
        if pb is not None:
            self.pbs[alias] = pb

    @property
    def pbs(self):
        """
        Get the probability dictionary of type :class:`~collections.OrderedDict`,
        which specifies the probabilities for the registered operators.
        """
        return self._pbs


__all__ = ['Toolbox']



# coding=utf-8
"""
.. moduleauthor:: Shuhua Gao

The module :mod:`mutation` provides mutation related genetic modification operators in GEP, including point-mutation,
transposition and inversion. Please refer to Chapter 3 of [FC2006]_ for more details.

.. note::
    1. Operators that can be applied to both :class:`~geppy.core.entity.Gene` and :class:`~geppy.core.entity.GeneDc`:
       :func:`mutate_uniform`, :func:`invert`, :func:`is_transpose`, :func:`ris_transpose`, :func:`gene_transpose`
    2. Operators that end with '_dc' are specially designed to handle the Dc domain in
       :class:`~geppy.core.entity.GeneDc`.
"""
import random
from ..core.symbol import Function, EphemeralTerminal
from ._util import _choose_a_terminal

_DEBUG = False


def _choose_function(pset):
    return random.choice(pset.functions)


def _choose_terminal(pset):
    return _choose_a_terminal(pset.terminals)


def _choose_subsequence(seq, min_length=1, max_length=-1):
    if max_length <= 0:
        max_length = len(seq)
    length = random.randint(min_length, max_length)
    start = random.randint(0, len(seq) - length)
    return start, start + length


def _choose_subsequence_indices(i, j, min_length=1, max_length=-1):
    """
    Choose a subsequence from [i, j] (both included) and return the subsequence boundaries [a, b] (both included).
    Additionally, the min_length and max length of [a, b], i.e., b - a + 1 can be specified.
    """
    if max_length <= 0:
        max_length = j - i + 1
    length = random.randint(min_length, max_length)
    start = random.randint(i, j - length + 1)
    return start, start + length - 1


def mutate_uniform(individual, pset, ind_pb='2p'):
    """
    Uniform point mutation. For each symbol (primitive) in *individual*, change it to another randomly chosen symbol
    from *pset* with the probability *indpb*. A symbol may be a function or a terminal.

    :param individual: :class:`~geppy.core.entity.Chromosome`, the chromosome to be mutated.
    :param pset: :class:`~geppy.core.entity.PrimitiveSet`, a primitive set
    :param ind_pb: float or str, default '2p'. Probability of mutating each symbol.
        If *ind_pb* is given as a string ending with 'p',
        then it indicates the expected number of point mutations among all the symbols in *individual*. For example,
        if the total length of each gene of *individual* is `l` and there are `m` genes in total, then by passing
        `ind_pb='1.5p'` we specify approximately `ind_pb=1.5/(l*m)`.
    :return: A tuple of one chromosome

    It is typical to set a mutation rate *indpb* equivalent to two one-point mutations per chromosome. That is,
    ``indpb = 2 / len(chromosome) * len(gene)``.
    """
    if isinstance(ind_pb, str):
        assert ind_pb.endswith('p'), "ind_pb must end with 'p' if given in a string form"
        length = individual[0].head_length + individual[0].tail_length
        ind_pb = float(ind_pb.rstrip('p')) / (len(individual) * length)
    for gene in individual:
        # mutate the gene with the associated pset
        # head: any symbol can be changed into a function or a terminal
        for i in range(gene.head_length):
            if random.random() < ind_pb:
                if random.random() < 0.5:  # to a function
                    gene[i] = _choose_function(pset)
                else:
                    gene[i] = _choose_terminal(pset)
        # tail: only change to another terminal
        for i in range(gene.head_length, gene.head_length + gene.tail_length):
            if random.random() < ind_pb:
                gene[i] = _choose_terminal(pset)
    return individual,


def mutate_uniform_dc(individual, ind_pb='1p'):
    """
    Dc-specific mutation. This operator changes one index stored in the Dc domain to another index in place. The indices
    in the Dc domain can later be used to retrieve numerical constants from the gene associated
    :meth:`~geppy.core.entity.GeneDc.rnc_array`.

    :param individual: :class:`~geppy.core.entity.Chromosome`, a chromosome, which contains genes of type
        :class:`~geppy.core.entity.GeneDc`
    :param ind_pb: float or str, default '1p'. Probability of mutating each index/position in the Dc domain. If given
        in str like 'xp', then `x` point mutations for expected for the Dc domains in total of *individual*. For
        instance, if there are `d` Dc elements in the whole chromosome, then `1.5p` is equal to `ind_pb = 1.5 / d`.
    :return: a tuple of one chromosome

    It is typical to set a mutation rate *indpb* equivalent to two one-point mutations per chromosome. That is,
    ``indpb = 2 / len(chromosome) * len(gene)``.
    """
    if isinstance(ind_pb, str):
        assert ind_pb.endswith('p'), "ind_pb must end with 'p' if given in a string form"
        ind_pb = float(ind_pb.rstrip('p')) / (individual[0].dc_length * len(individual))
    # change the position in the Dc domain to another one
    for g in individual:  # g: GeneDc
        start = g.head_length + g.tail_length
        end = start + g.dc_length
        for i in range(start, end):
            if random.random() < ind_pb:
                g[i] = random.randint(0, len(g.rnc_array) - 1)
    return individual,


def invert(individual):
    """
    A gene is randomly chosen, and afterwards a subsequence within this gene's head domain is randomly selected
    and inverted.

    :param individual: :class:`~geppy.core.entity.Chromosome`, a chromosome
    :return: a tuple of one individual

    Typically, a small inversion rate of 0.1 is used.
    """
    if individual.head_length < 2:
        return individual,
    gene = random.choice(individual)
    start, end = _choose_subsequence(gene.head, 2, gene.head_length)
    gene[start: end] = reversed(gene[start: end])
    if _DEBUG:
        print('invert [{}: {}]'.format(start, end))
    return individual,


def invert_dc(individual):
    """
    Dc-specific inversion. A gene is randomly chosen, and afterwards a subsequence within this gene's Dc domain is
    randomly selected and inverted.

    :param individual: :class:`~geppy.core.entity.Chromosome`, a chromosome, which contains genes of type
        :class:`~geppy.core.entity.GeneDc`
    :return: a tuple of one individual

    Typically, a small inversion rate of 0.1 is used.
    """
    g = random.choice(individual)
    if g.dc_length < 2:
        return individual,
    start = g.head_length + g.tail_length
    end = start + g.dc_length - 1
    inv_start, inv_end = _choose_subsequence_indices(start, end, min_length=2)
    g[inv_start: inv_end] = reversed(g[inv_start: inv_end])
    return individual,


def _choose_donor_donee(individual):
    i1, i2 = random.choices(range(len(individual)), k=2)  # with replacement
    return individual[i1], individual[i2], i1, i2


def is_transpose(individual):
    """
    Perform IS transposition in place

    :param individual: :class:`~geppy.core.entity.Chromosome`, a chromosome
    :return: a tuple of one individual

    An Insertion Sequence (IS) is a random chosen segment across the chromosome, and then the
    IS is copied to be inserted at another position in the head of gene, except the start position. Note that an IS from
    one gene may be copied to a different gene.
    Typically, the IS transposition rate is set around 0.1.
    """
    donor, donee, i1, i2 = _choose_donor_donee(individual)
    a, b = _choose_subsequence_indices(0, donor.head_length + donor.tail_length - 1, max_length=donee.head_length - 1)
    is_start, is_end = a, b + 1
    is_ = donor[is_start: is_end]
    insertion_pos = random.randint(1, donee.head_length - len(is_))
    donee[:] = donee[:insertion_pos] + is_ + \
               donee[insertion_pos: insertion_pos + donee.head_length - insertion_pos - len(is_)] + \
               donee[donee.head_length:]
    if _DEBUG:
        print('IS transpose: g{}[{}:{}] -> g{}[{}:]'.format(i1, is_start, is_end, i2, insertion_pos))
    return individual,


def ris_transpose(individual):
    """
    Perform RIS transposition in place

    :param individual: :class:`~geppy.core.entity.Chromosome`, a chromosome
    :return: a tuple of one individual

    A Root Insertion Sequence (RIS) is a segment of consecutive elements that starts with a
    function. Once an RIS is randomly chosen, it is copied and inserted at the root (first position) of a gene. Note
    that an RIS from one gene may be copied to a different gene.
    Typically, the RIS transposition rate is set around 0.1.
    """
    n_trial = 0
    while n_trial <= 2 * len(individual):
        donor, donee, i1, i2 = _choose_donor_donee(individual)
        # choose a function node randomly to start RIS
        function_indices = [i for i, p in enumerate(donor.head) if isinstance(p, Function)]
        if not function_indices:  # no functions in this donor, try another
            n_trial += 1
            continue
        ris_start = random.choice(function_indices)
        # determine the length randomly
        length = random.randint(2, min(donee.head_length, donor.head_length + donor.tail_length - ris_start))
        # insert ris at the root of donee
        ris = donor[ris_start: ris_start + length]
        donee[:] = ris + donee[0: donee.head_length - length] + donee[donee.head_length:]
        if _DEBUG:
            print('RIS transpose: g{}[{}:{}] -> g{}[0:]'.format(i1, ris_start, ris_start + length, i2))
        return individual,
    return individual,


def gene_transpose(individual):
    """
    Perform gene transposition in place

    :param individual: :class:`~geppy.core.entity.Chromosome`, a chromosome
    :return: a tuple of one individual

    An entire gene is selected randomly and exchanged with the first gene in the chromosome.
    Obviously, this operation only makes sense if the chromosome is a multigenic one. Typically, the gene transposition
    rate is set around 0.1.

    """
    if len(individual) <= 1:
        return individual,
    source = random.randint(1, len(individual) - 1)
    individual[0], individual[source] = individual[source], individual[0]
    if _DEBUG:
        print('Gene transpose: g0 <-> g{}'.format(source))
    return individual,


def transpose_dc(individual):
    """
    This operator randomly chooses the chromosome, the gene with its respective Dc to be subjected to transposition,
    the start and termination points of the transposon, and the target site; then it inserts a copy of the transposon
    from the place of origin to the target site, while the old content at the target site are shifted to the right
    and the length of the Dc domain is maintained.

    :param individual: :class:`~geppy.core.entity.Chromosome`, a chromosome, which contains genes of type
        :class:`~geppy.core.entity.GeneDc`
    :return: a tuple of one individual

    Typically, a small transposition rate of 0.1 is used. Note that a transposon from one gene's Dc may be copied and
    inserted into another gene's Dc, which means the genes in a multigenic chromosome *individual*
    must have an identical structure.
    """
    donor, donee, i, j = _choose_donor_donee(individual)
    dc_start = donor.head_length + donor.tail_length
    dc_end = donor.head_length + donor.tail_length + donor.dc_length - 1  # included
    # choose a transposon from the donor Dc
    t_start, t_end = _choose_subsequence_indices(dc_start, dc_end)
    # choose the insertion point in donee Dc
    insertion_pos = random.randint(dc_start, dc_end - (t_end - t_start))
    donee[:] = donee[:insertion_pos] + donor[t_start: t_end + 1] \
               + donee[insertion_pos: insertion_pos + len(donee) - (insertion_pos + t_end - t_start + 1)]
    if _DEBUG:
        print('Dc transpose: g{}[{}:{}] -> g{}[{}] (both included)'.format(i, t_start, t_end, j, insertion_pos))
    return individual,


def mutate_rnc_array_dc(individual, rnc_gen, ind_pb='1p'):
    """
    Direct mutation of RNCs, which changes the values in a gene's RNC array
    :meth:`~geppy.core.entity.GeneDc.rnc_array` randomly.

    :param individual: :class:`~geppy.core.entity.Chromosome`, a chromosome, which contains genes of type
        :class:`~geppy.core.entity.GeneDc`.
    :param rnc_gen: callable, which returns a random numerical constant by calling ``rnc_gen()``.
    :param ind_pb: float or str, default '1p'. Use a float number to specify the probability of each RNC in the array
        being mutated. Alternatively, if a str ending with 'p' is given in the form like 'xp',  where
        'x' is a numerical number, then it specifies how many point mutations are expected for the RNC arrays
        in *individual*. For example, if there are *d* RNCs inside the arrays in total in *individual*,
        then '1.5p' is approximately equal to `ind_pb = 1.5 / d`.
    :return: a tuple of one individual

    The genetic operators :func:`mutate_uniform_dc`, :func:`transpose_dc` and :func:`invert_dc` actually only move the
    random numerical constants around without generating new numerical values. This method :func:`mutate_rnc_array_dc`
    can replace the value of a particular numerical constant by another in the
    :meth:`~geppy.core.entity.GeneDc.rnc_array`.

    Refer to section 5.4.4 of [FC2006]_ for more details.
    """
    if isinstance(ind_pb, str):
        assert ind_pb.endswith('p'), "ind_pb must end with 'p' if given in a string form"
        n_rnc_total = len(individual[0].rnc_array) * len(individual)
        if n_rnc_total == 0:
            return individual,
        ind_pb = float(ind_pb.rstrip('p')) / n_rnc_total

    for g in individual:
        for i in range(len(g.rnc_array)):
            if random.random() < ind_pb:
                g.rnc_array[i] = rnc_gen()
    return individual,


def mutate_uniform_ephemeral(individual, ind_pb='1p'):
    """
    This operator is specially designed for ephemeral numerical constant (ENC) mutation.
    It attempts to change the value of the ephemeral constants in the *individual* if any by calling `e.update_value()`,
    where `e` is an ephemeral inside *individual*. See :meth:`~geppy.core.symbol.Ephemeral.update_value` for details.

    :param individual: :class:`~geppy.core.entity.Chromosome`, a chromosome
    :param ind_pb: float or str, default '1p'. If a float is given, then *ind_pb* specifies the probability of each
        ephemeral in the *individual* being mutated. If a str ending with 'p' is given in the form like 'xp',  where
        'x' is a numerical number, then it
        specifies how many point mutations are expected for the ephemeral constants. For example, if there are *d*
        ephemeral constants present in total in *individual*, then '1.5p' is approximately equal to `ind_pb = 1.5 / d`.

    :return: a tuple of one individual
    """
    ephemerals = [p for g in individual for p in g if isinstance(p, EphemeralTerminal)]
    if len(ephemerals) == 0:
        return individual,
    if isinstance(ind_pb, str):
        assert ind_pb.endswith('p'), "ind_pb must end with 'p' if given in a string form"
        np = float(ind_pb.rstrip('p'))
        ind_pb = np / len(ephemerals)
    for e in ephemerals:
        if random.random() < ind_pb:
            e.update_value()
    return individual,



# coding=utf-8
"""
.. moduleauthor:: Shuhua Gao

This module :mod:`generator` provides functionality to generate a genome for a gene. That is, choose functions and
terminals randomly from a given primitive set to form a linear form gene expression.
"""
import random
from ._util import _choose_a_terminal


def generate_genome(pset, head_length):
    """
    Generate a genome with the given primitive set *pset* and the specified head domain length *head_length*.

    :param pset: a primitive set
    :param head_length: length of the head domain
    :return: a list of symbols representing a genome

    Supposing the maximum arity of functions in *pset* is *max_arity*, then the tail length is automatically
    determined to be ``tail_length = head_length * (max_arity - 1) + 1``.
    """
    h = head_length
    functions = pset.functions
    terminals = pset.terminals

    n_max = max(p.arity for p in functions)  # max arity
    t = h * (n_max - 1) + 1
    expr = [None] * (h + t)
    # head part: initialized with both functions and terminals
    for i in range(h):
        if random.random() < 0.5:
            expr[i] = random.choice(functions)
        else:
            expr[i] = _choose_a_terminal(terminals)
    # tail part: only terminals are allowed
    for i in range(h, h + t):
        expr[i] = _choose_a_terminal(terminals)
    return expr


def generate_dc(rnc_array_length, dc_length):
    """
    Generate a Dc domain for a RNC array of size *rnc_array_length*.

    :param rnc_array_length: length of the RNC array
    :param dc_length: length of the Dc domain
    :return: a list of integers of length *dc_length*, each element in range `[0, rnc_array_length - 1]`

    Refer to Chapter 5 of [FC2006]_ for more details.
    """
    return [random.randint(0, rnc_array_length - 1) for _ in range(dc_length)]




"""
Some utility functions used internally.
"""

import random
import copy
from ..core.symbol import EphemeralTerminal


def _choose_a_terminal(terminals):
    """
    Choose a terminal from the given list *terminals* randomly.

    :param terminals: iterable, terminal set
    :return: a terminal
    """
    terminal = random.choice(terminals)
    if isinstance(terminal, EphemeralTerminal):  # an Ephemeral
        terminal = copy.deepcopy(terminal)  # generate a new one
        terminal.update_value()
    return terminal



# coding=utf-8
"""
.. moduleauthor:: Shuhua Gao

This module :mod:`simplification` provides utility functions for symbolic simplification of GEP individuals, which may
be used in postprocessing.
"""
import math
import operator
from ..core.entity import KExpression, Chromosome, Gene
from ..core.symbol import Function, Terminal, SymbolTerminal

import sympy as sp


DEFAULT_SYMBOLIC_FUNCTION_MAP = {
    operator.and_.__name__: sp.And,
    operator.or_.__name__: sp.Or,
    operator.not_.__name__: sp.Not,
    operator.add.__name__: operator.add,
    operator.sub.__name__: operator.sub,
    operator.mul.__name__: operator.mul,
    operator.neg.__name__: operator.neg,
    operator.pow.__name__: operator.pow,
    operator.abs.__name__: operator.abs,
    operator.floordiv.__name__: operator.floordiv,
    operator.truediv.__name__: operator.truediv,
    'protected_div': operator.truediv,
    math.log.__name__: sp.log,
    math.sin.__name__: sp.sin,
    math.cos.__name__: sp.cos,
    math.tan.__name__: sp.tan
}
"""
Currently, it is defined as::

    DEFAULT_SYMBOLIC_FUNCTION_MAP = {
        operator.and_.__name__: sp.And,
        operator.or_.__name__: sp.Or,
        operator.not_.__name__: sp.Not,
        operator.add.__name__: operator.add,
        operator.sub.__name__: operator.sub,
        operator.mul.__name__: operator.mul,
        operator.neg.__name__: operator.neg,
        operator.pow.__name__: operator.pow,
        operator.abs.__name__: operator.abs,
        operator.floordiv.__name__: operator.floordiv,
        operator.truediv.__name__: operator.truediv,
        'protected_div': operator.truediv,
        math.log.__name__: sp.log,
        math.sin.__name__: sp.sin,
        math.cos.__name__: sp.cos,
        math.tan.__name__: sp.tan
    }
"""


def _simplify_kexpression(expr, symbolic_function_map):
    """
    Simplify a K-expression.
    :return: a symbolic expression
    """
    assert len(expr) > 0
    if len(expr) == 1:  # must be a single terminal
        t = expr[0]
        assert isinstance(
            t, Terminal), 'A K-expression of length 1 must only contain a terminal.'
        if t.value is None:  # an input
            return sp.Symbol(t.name)
        return t.value

    expr = expr[:]  # because we need to change expr
    # K-expression is simply a level-order serialization of an expression tree.
    for i in reversed(range(len(expr))):
        p = expr[i]
        if isinstance(p, Function):
            try:
                sym_func = symbolic_function_map[p.name]
            except KeyError as e:
                print(
                    "Please provide the symbolic function mapping for '{}' in symbolic_function_map.".format(p.name))
                raise e
            args = []
            for _ in range(p.arity):
                t = expr.pop()  # t may be a terminal or a symbolic expression already
                if isinstance(t, Terminal):
                    if isinstance(t, SymbolTerminal):
                        args.append(sp.Symbol(t.name))
                    else:
                        args.append(t.value)
                else:
                    args.append(t)
            # evaluate this function node symbolically
            r = sym_func(*reversed(args))
            expr[i] = sp.simplify(r)
    return expr[0]


def simplify(genome, symbolic_function_map=None):
    """
    Compile the primitive tree into a (possibly simplified) symbolic expression.

    :param genome: :class:`~geppy.core.entity.KExpression`, :class:`~geppy.core.entity.Gene`, or
        :class:`~geppy.core.entity.Chromosome`, the genotype of an individual
    :param symbolic_function_map: dict, maps each function name in the primitive set to a symbolic version
    :return: a (simplified) symbol expression

    For example, *add(sub(3, 3), x)* may be simplified to *x*. This :func:`simplify` function can be used to
    postprocess the best individual obtained in GEP for a simplified representation. Some Python functions like
    :func:`operator.add` can be used directly in *sympy*. However, there are also functions that have their own
    symbolic versions to be used in *sympy*, like the :func:`operator.and_`, which should be replaced by
    :func:`sympy.And`. In such a case, we may provide a map
    ``symbolic_function_map={operator.and_.__name__, sympy.And}`` supposing the function primitive encapsulating
    :func:`operator.and_` uses its default name.

    Such simplification doesn't affect GEP at all. It should be used as a postprocessing step to simplify the final
    solution evolved by GEP.

    .. note::
        If the *symbolic_function_map* argument remains as the default value ``None``, then a default map
        :data:`DEFAULT_SYMBOLIC_FUNCTION_MAP` is used, which contains common
        *name-to-symbolic function* mappings, including the arithmetic operators and Boolean logic operators..

    .. note::
        This function depends on the :mod:`sympy` module. You can find it `here <http://www.sympy.org/en/index.html>`_.
    """
    if symbolic_function_map is None:
        symbolic_function_map = DEFAULT_SYMBOLIC_FUNCTION_MAP
    if isinstance(genome, KExpression):
        return _simplify_kexpression(genome, symbolic_function_map)
    elif isinstance(genome, Gene):
        return _simplify_kexpression(genome.kexpression, symbolic_function_map)
    elif isinstance(genome, Chromosome):
        if len(genome) == 1:
            return _simplify_kexpression(genome[0].kexpression, symbolic_function_map)
        else:   # multigenic chromosome
            simplified_exprs = [_simplify_kexpression(
                g.kexpression, symbolic_function_map) for g in genome]
            # combine these sub-expressions into a single one with the linking function
            try:
                linker = symbolic_function_map[genome.linker.__name__]
            except:
                linker = genome.linker
            return sp.simplify(linker(*simplified_exprs))
    else:
        raise TypeError('Only an argument of type KExpression, Gene, and Chromosome is acceptable. The provided '
                        'genome type is {}.'.format(type(genome)))



# coding=utf-8
"""
The *geppy* package supports gene expression programming in Python.

.. tip::
    For convenience of usage, all APIs have been imported into the top-level package *geppy* and can be used with the
    short form ``geppy.API`` directly. For example, simply use ``geppy.Chromosome`` as an alias of the class
    :class:`geppy.core.entity.Chromosome` and ``geppy.invert`` as the alias of the function
    :func:`geppy.tools.mutation.invert`.
"""
#    This file is part of geppy.
#
#    geppy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    geppy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with geppy. If not, see <http://www.gnu.org/licenses/>.

from pkg_resources import get_distribution, DistributionNotFound

# fetch version from setup.py
try:
    __version__ = get_distribution('geppy').version
except DistributionNotFound as e:
    __version__ = 'Please install this package with setup.py'

__author__ = 'Shuhua Gao'
# __revision__ = "1.2.2"

from .core.entity import *
from .core.symbol import *
from .tools.crossover import *
from .tools.mutation import *
from .algorithms.basic import *
from .tools.parser import *
from .tools.generator import *
from .support.visualization import *
from .tools.toolbox import *


def _print_module_not_found(module_name, f_name, http=''):
    print('''The module {0} is not available, on which {1} depends.
                The geepy.graph function will not be visible until this module is installed. 
                To install graphviz, please check {2}.'''.format(module_name, f_name, http))


try:
    import graphviz
    from geppy.support.visualization import export_expression_tree
except ImportError:
    _print_module_not_found('graphviz', 'geepy.export_expression_tree', 'https://pypi.org/project/graphviz/')

try:
    import sympy
    from geppy.support.simplification import simplify
except ImportError:
    _print_module_not_found('sympy', 'geepy.simplify', 'http://www.sympy.org/')





# coding=utf-8
"""
.. moduleauthor:: Shuhua Gao

This module :mod:`parser` provides functionality for compiling an individual (a chromosome) in GEP into an executable
lambda function in Python for subsequent fitness evaluation.

.. todo::
    Parse an individual into codes of other languages, such as C++/Java, for deployment in an industrial environment.
"""

import sys


def _compile_gene(g, pset):
    """
    Compile one gene *g* with the primitive set *pset*.
    :return: a function or an evaluated result
    """
    code = str(g)
    if len(pset.input_names) > 0:   # form a Lambda function
        args = ', '.join(pset.input_names)
        code =  'lambda {}: {}'.format(args, code)
    # evaluate the code
    try:
        return eval(code, pset.globals, {})
    except MemoryError:
        _, _, traceback = sys.exc_info()
        raise MemoryError("The expression tree generated by GEP is too deep. Python cannot evaluate a tree higher "
                          "than 90. You should try to adopt a smaller head length for the genes, for example, by using"
                          "more genes in a chromosome.").with_traceback(traceback)


def compile_(individual, pset):
    """
    Compile the individual into a Python lambda expression.

    :param individual: :class:`Chromosome`, a chromosome
    :param pset: :class:`PrimitiveSet`, a primitive set
    :return: a function if the primitive set *pset* has any inputs (arguments), which can later be called with
        specific parameter values; otherwise, a numerical result obtained from evaluation.
    """
    fs = [_compile_gene(gene, pset) for gene in individual]
    linker = individual.linker
    if linker is None:
        if len(fs) == 1:
            return fs[0]
        else:
            return lambda *x: tuple((f(*x) for f in fs))
    return lambda *x: linker(*(f(*x) for f in fs))


__all__ = ['compile_']


# coding=utf-8
import setuptools

try:
    with open("README.md", "r") as fh:
        long_description = fh.read()
except:
    long_description = '''*geppy* is developed for fast prototyping of **gene expression programming** (GEP) in Python, which is
    built on top of the DEAP package. Please the [GitHub repository](https://github.com/ShuhuaGao/geppy) for details.'''

setuptools.setup(
    name="geppy",
    version="0.1",
    license='LGPL-3.0 License',
    author="Shuhua Gao",
    author_email="nus.gao.shuhua@gmail.com",
    description="A package for gene expression programming in Python.",
    long_description=long_description,
    url="https://github.com/ShuhuaGao/geppy",
    classifiers=(
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ),
	long_description_content_type='text/markdown',
    keywords=['evolutionary computation', 'gene expression programming', 'computational intelligence'],
    packages=setuptools.find_packages(),
    install_requires=['deap'],
    download_url='https://github.com/ShuhuaGao/geppy/archive/0.1.tar.gz'
)




# coding=utf-8
"""
.. moduleauthor:: Shuhua Gao

The module :mod:`crossover` provides crossover (mating, recombination) related genetic modifications in GEP,
including one-point and two-point crossover, and gene crossover between multigenic chromosomes.
Please refer to Chapter 3 of [FC2006]_ for more details.

.. note::
    All the recombination operators including :func:`crossover_one_point`, :func:`crossover_two_point`, and
    :func:`crossover_gene` can be applied to both :class:`~geppy.core.entity.Gene`
    and :class:`~geppy.core.entity.GeneDc`.

.. [FC2006] Ferreira, Cândida. Gene expression programming: mathematical modeling by an artificial
    intelligence. Vol. 21. Springer, 2006.

"""
import random

_DEBUG = False


def crossover_one_point(ind1, ind2):
    """
    Execute one-point recombination of two individuals. The two individuals are modified in place, and the two children
    are returned.

    :param ind1: The first individual (chromosome) participating in the crossover.
    :param ind2: The second individual (chromosome) participating in the crossover.
    :return: A tuple of two children individuals.

    Note the crossover can happen at any point across the whole chromosome and thus entire genes may be also exchanged
    between the two parents if they are multigenic chromosomes.
    """
    assert len(ind1) == len(ind2)
    # the gene containing the recombination point, and the point index in the gene
    which_gene = random.randint(0, len(ind1) - 1)
    which_point = random.randint(0, len(ind1[which_gene]) - 1)
    # exchange the upstream materials
    ind1[:which_gene], ind2[:which_gene] = ind2[:which_gene], ind1[:which_gene]
    ind1[which_gene][:which_point + 1], ind2[which_gene][:which_point + 1] = \
        ind2[which_gene][:which_point + 1], ind1[which_gene][:which_point + 1]
    if _DEBUG:
        print('cxOnePoint: g{}[{}]'.format(which_gene, which_point))
    return ind1, ind2


def crossover_two_point(ind1, ind2):
    """
    Execute two-point recombination of two individuals. The two individuals are modified in place, and the two children
    are returned. The materials between two randomly chosen points are swapped to generate two children.

    :param ind1: The first individual (chromosome) participating in the crossover.
    :param ind2: The second individual (chromosome) participating in the crossover.
    :return: A tuple of two individuals.

    Note the crossover can happen at any point across the whole chromosome and thus entire genes may be also exchanged
    between the two parents if they are multigenic chromosomes.
    """
    assert len(ind1) == len(ind2)
    # the two genes containing the two recombination points
    g1, g2 = random.choices(range(len(ind1)), k=2)  # with replacement, thus g1 may be equal to g2
    if g2 < g1:
        g1, g2 = g2, g1
    # the two points in g1 and g2
    p1 = random.randint(0, len(ind1[g1]) - 1)
    p2 = random.randint(0, len(ind1[g2]) - 1)
    # change the materials between g1->p1 and g2->p2: first exchange entire genes, then change partial genes at g1, g2
    if g1 == g2:
        if p1 > p2:
            p1, p2 = p2, p1
        ind1[g1][p1: p2+1], ind2[g2][p1: p2+1] = ind2[g2][p1: p2+1], ind1[g1][p1: p2+1]
    else:
        ind1[g1 + 1: g2], ind2[g1 + 1: g2] = ind2[g1 + 1: g2], ind1[g1 + 1: g2]
        ind1[g1][p1:], ind2[g1][p1:] = ind2[g1][p1:], ind1[g1][p1:]
        ind1[g2][:p2 + 1], ind2[g2][:p2 + 1] = ind2[g2][:p2 + 1], ind1[g2][:p2 + 1]
    if _DEBUG:
        print('cxTwoPoint: g{}[{}], g{}[{}]'.format(g1, p1, g2, p2))
    return ind1, ind2


def crossover_gene(ind1, ind2):
    """
    Entire genes are exchanged between two parent chromosomes. The two individuals are modified in place, and the two
    children are returned.

    :param ind1: The first individual (chromosome) participating in the crossover.
    :param ind2: The second individual (chromosome) participating in the crossover.
    :return: a tuple of two children individuals

    This operation has no effect if the chromosome has only one gene. Typically, a gene recombination rate
    around 0.2 is used.
    """
    assert len(ind1) == len(ind2)
    pos1, pos2 = random.choices(range(len(ind1)), k=2)
    ind1[pos1], ind2[pos2] = ind2[pos2], ind1[pos1]
    if _DEBUG:
        print('cxGene: ind1[{}] <--> ind2[{}]'.format(pos1, pos2))
    return ind1, ind2



# coding=utf-8
"""
.. moduleauthor:: Shuhua Gao

This :mod:`basic` module provides fundamental boilerplate GEP algorithm implementations. After registering proper
operations into a :class:`deap.base.Toolbox` object, the GEP evolution can be simply launched using the present
algorithms. Of course, for complicated problems, you may want to define your own algorithms, and the implementation here
can be used as a reference.
"""
import deap
import random
import warnings


def _validate_basic_toolbox(tb):
    """
    Validate the operators in the toolbox *tb* according to our conventions.
    """
    assert hasattr(tb, 'select'), "The toolbox must have a 'select' operator."
    # whether the ops in .pbs are all registered
    for op in tb.pbs:
        assert op.startswith('mut') or op.startswith('cx'), "Operators must start with 'mut' or 'cx' except selection."
        assert hasattr(tb, op), "Probability for a operator called '{}' is specified, but this operator is not " \
                                "registered in the toolbox.".format(op)
    # whether all the mut_ and cx_ operators have their probabilities assigned in .pbs
    for op in [attr for attr in dir(tb) if attr.startswith('mut') or attr.startswith('cx')]:
        if op not in tb.pbs:
            warnings.warn('{0} is registered, but its probability is NOT assigned in Toolbox.pbs. '
                          'By default, the probability is ZERO and the operator {0} will NOT be applied.'.format(op),
                          category=UserWarning)


def _apply_modification(population, operator, pb):
    """
    Apply the modification given by *operator* to each individual in *population* with probability *pb* in place.
    """
    for i in range(len(population)):
        if random.random() < pb:
            population[i], = operator(population[i])
            del population[i].fitness.values
    return population


def _apply_crossover(population, operator, pb):
    """
    Mate the *population* in place using *operator* with probability *pb*.
    """
    for i in range(1, len(population), 2):
        if random.random() < pb:
            population[i - 1], population[i] = operator(population[i - 1], population[i])
            del population[i - 1].fitness.values
            del population[i].fitness.values
    return population


def gep_simple(population, toolbox, n_generations=100, n_elites=1,
               stats=None, hall_of_fame=None, verbose=__debug__):
    """
    This algorithm performs the simplest and standard gene expression programming.
    The flowchart of this algorithm can be found
    `here <https://www.gepsoft.com/gxpt4kb/Chapter06/Section1/SS1.htm>`_.
    Refer to Chapter 3 of [FC2006]_ to learn more about this basic algorithm.

    .. note::
        The algorithm framework also supports the GEP-RNC algorithm, which evolves genes with an additional Dc domain for
        random numerical constant manipulation. To adopt :func:`gep_simple` for GEP-RNC evolution, use the
        :class:`~geppy.core.entity.GeneDc` objects as the genes and register Dc-specific operators.
        A detailed example of GEP-RNC can be found at `numerical expression inference with GEP-RNC
        <https://github.com/ShuhuaGao/geppy/blob/master/examples/sr/numerical_expression_inference-RNC.ipynb>`_.
        Users can refer to Chapter 5 of [FC2006]_ to get familiar with the GEP-RNC theory.

    :param population: a list of individuals
    :param toolbox: :class:`~geppy.tools.toolbox.Toolbox`, a container of operators. Regarding the conventions of
        operator design and registration, please refer to :ref:`convention`.
    :param n_generations: max number of generations to be evolved
    :param n_elites: number of elites to be cloned to next generation
    :param stats: a :class:`~deap.tools.Statistics` object that is updated
                  inplace, optional.
    :param hall_of_fame: a :class:`~deap.tools.HallOfFame` object that will
                       contain the best individuals, optional.
    :param verbose: whether or not to print the statistics.
    :returns: The final population
    :returns: A :class:`~deap.tools.Logbook` recording the statistics of the
              evolution process

    .. note:
        To implement the GEP-RNC algorithm for numerical constant evolution, the :class:`geppy.core.entity.GeneDc` genes
        should be used. Specific operators are used to evolve the Dc domain of :class:`~geppy.core.entity.GeneDc` genes
        including Dc-specific mutation/inversion/transposition and direct mutation of the RNC array associated with
        each gene. These operators should be registered into the *toolbox*.
    """
    _validate_basic_toolbox(toolbox)
    logbook = deap.tools.Logbook()
    logbook.header = ['gen', 'nevals'] + (stats.fields if stats else [])

    for gen in range(n_generations + 1):
        # evaluate: only evaluate the invalid ones, i.e., no need to reevaluate the unchanged ones
        invalid_individuals = [ind for ind in population if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_individuals)
        for ind, fit in zip(invalid_individuals, fitnesses):
            ind.fitness.values = fit

        # record statistics and log
        if hall_of_fame is not None:
            hall_of_fame.update(population)
        record = stats.compile(population) if stats else {}
        logbook.record(gen=gen, nevals=len(invalid_individuals), **record)
        if verbose:
            print(logbook.stream)

        if gen == n_generations:
            break

        # selection with elitism
        elites = deap.tools.selBest(population, k=n_elites)
        offspring = toolbox.select(population, len(population) - n_elites)

        # replication
        offspring = [toolbox.clone(ind) for ind in offspring]

        # mutation
        for op in toolbox.pbs:
            if op.startswith('mut'):
                offspring = _apply_modification(offspring, getattr(toolbox, op), toolbox.pbs[op])

        # crossover
        for op in toolbox.pbs:
            if op.startswith('cx'):
                offspring = _apply_crossover(offspring, getattr(toolbox, op), toolbox.pbs[op])

        # replace the current population with the offsprings
        population = elites + offspring

    return population, logbook


__all__ = ['gep_simple']


